# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from .interface import SchedulerInterface
from .model import ListedOrigin, Lister


def utcnow():
    return datetime.now(tz=timezone.utc)


def get_task(task_name):
    """Retrieve task object in our application instance by its fully
    qualified python name.

    Args:
        task_name (str): task's name (e.g
                         swh.loader.git.tasks.LoadDiskGitRepository)

    Returns:
        Instance of task

    """
    from swh.scheduler.celery_backend.config import app

    for module in app.conf.CELERY_IMPORTS:
        __import__(module)
    return app.tasks[task_name]


def create_task_dict(
    type: str, policy: str, *args, next_run: Optional[datetime] = None, **kwargs
) -> Dict[str, Any]:
    """Create a task with type and policy, scheduled for as soon as
       possible.

    Args:
        type: Type of oneshot task as per swh-scheduler's db
                    table task_type's column (Ex: load-git,
                    check-deposit)
        policy: oneshot or recurring policy
        next_run: optional date and time from which the task can be executed,
            use current time otherwise

    Returns:
        Expected dictionary for the one-shot task scheduling api
        (swh.scheduler.backend.create_tasks)

    """
    task_extra = {}
    for extra_key in ["priority", "retries_left"]:
        if extra_key in kwargs:
            extra_val = kwargs.pop(extra_key)
            task_extra[extra_key] = extra_val

    task = {
        "policy": policy,
        "type": type,
        "next_run": next_run or utcnow(),
        "arguments": {
            "args": args if args else [],
            "kwargs": kwargs if kwargs else {},
        },
    }
    task.update(task_extra)
    return task


def create_origin_task_dict(origin: ListedOrigin, lister: Lister) -> Dict[str, Any]:
    if origin.lister_id != lister.id:
        raise ValueError(
            "origin.lister_id and lister.id differ", origin.lister_id, lister.id
        )
    return {
        "type": f"load-{origin.visit_type}",
        "arguments": {
            "args": [],
            "kwargs": {
                "url": origin.url,
                "lister_name": lister.name,
                "lister_instance_name": lister.instance_name or None,
                **origin.extra_loader_arguments,
            },
        },
    }


def create_origin_task_dicts(
    origins: List[ListedOrigin], scheduler: SchedulerInterface
) -> List[Dict[str, Any]]:
    """Returns a task dict for each origin, in the same order."""

    lister_ids = {o.lister_id for o in origins}
    listers = {
        lister.id: lister
        for lister in scheduler.get_listers_by_id(list(map(str, lister_ids)))
    }

    missing_lister_ids = lister_ids - set(listers)
    assert not missing_lister_ids, f"Missing listers: {missing_lister_ids}"

    return [create_origin_task_dict(o, listers[o.lister_id]) for o in origins]


def create_oneshot_task_dict(
    type: str, *args, next_run: Optional[datetime] = None, **kwargs
):
    """Create a oneshot task scheduled for as soon as possible.

    Args:
        type: Type of oneshot task as per swh-scheduler's db
            table task_type's column (Ex: load-git, check-deposit)
        next_run: optional date and time from which the task can be executed,
            use current time otherwise

    Returns:
        Expected dictionary for the one-shot task scheduling api
        (:func:`swh.scheduler.backend.create_tasks`)

    """
    return create_task_dict(type, "oneshot", *args, next_run=next_run, **kwargs)



# Copyright (C) 2021-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import logging

import click

from . import cli as cli_scheduler_group

logger = logging.getLogger(__name__)


@cli_scheduler_group.command("journal-client")
@click.pass_context
@click.option(
    "--stop-after-objects",
    "-m",
    default=None,
    type=int,
    help="Maximum number of objects to replay. Default is to run forever.",
)
def visit_stats_journal_client(ctx, stop_after_objects):
    """Keep the the origin visits stats table up to date from a swh kafka journal

    Expected configuration:

    \b
    * :ref:`cli-config-journal`
    * :ref:`cli-config-scheduler`
    """
    from functools import partial

    from swh.journal.client import get_journal_client
    from swh.scheduler.journal_client import process_journal_objects

    if not ctx.obj["scheduler"]:
        message = "Scheduler class (local/remote) must be instantiated."
        any_exception = ctx.obj.get("scheduler_exc")
        if any_exception:
            extra_message = f"Scheduler problems: {any_exception}"
            message = "\n".join([message, extra_message])
            logger.exception(any_exception)
        ctx.fail(message)

    scheduler = ctx.obj["scheduler"]
    config = ctx.obj["config"]

    if "journal" not in config:
        ctx.fail("Missing 'journal' configuration key")

    journal_cfg = config["journal"]
    journal_cfg["stop_after_objects"] = stop_after_objects or journal_cfg.get(
        "stop_after_objects"
    )

    client = get_journal_client(
        cls="kafka",
        object_types=["origin_visit_status"],
        prefix="swh.journal.objects",
        **journal_cfg,
    )
    worker_fn = partial(
        process_journal_objects,
        scheduler=scheduler,
    )
    nb_messages = 0
    try:
        nb_messages = client.process(worker_fn)
        print(f"Processed {nb_messages} message(s).")
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        print("Done.")
    finally:
        client.close()



# Copyright (C) 2021-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import pytest
import yaml

from swh.scheduler.celery_backend.config import (
    MAX_NUM_TASKS,
    app,
    get_available_slots,
    route_for_task,
    setup_log_handler,
)


@pytest.mark.parametrize("name", ["swh.something", "swh.anything"])
def test_route_for_task_routing(name):
    assert route_for_task(name, [], {}, {}) == {"queue": name}


@pytest.mark.parametrize("name", [None, "foobar"])
def test_route_for_task_no_routing(name):
    assert route_for_task(name, [], {}, {}) is None


def test_get_available_slots_no_max_length():
    actual_num = get_available_slots(app, "anything", None)
    assert actual_num == MAX_NUM_TASKS


def test_get_available_slots_issue_when_reading_queue(mocker):
    mock = mocker.patch("swh.scheduler.celery_backend.config.get_queue_length")
    mock.side_effect = ValueError

    actual_num = get_available_slots(app, "anything", max_length=10)
    assert actual_num == MAX_NUM_TASKS
    assert mock.called


def test_get_available_slots_no_queue_length(mocker):
    mock = mocker.patch("swh.scheduler.celery_backend.config.get_queue_length")
    mock.return_value = None
    actual_num = get_available_slots(app, "anything", max_length=100)
    assert actual_num == MAX_NUM_TASKS
    assert mock.called


def test_get_available_slots_no_more_slots(mocker):
    mock = mocker.patch("swh.scheduler.celery_backend.config.get_queue_length")
    max_length = 100
    queue_length = 9000
    mock.return_value = queue_length
    actual_num = get_available_slots(app, "anything", max_length)
    assert actual_num == 0
    assert mock.called


def test_get_available_slots(mocker):
    mock = mocker.patch("swh.scheduler.celery_backend.config.get_queue_length")
    max_length = 100
    queue_length = 90
    mock.return_value = queue_length
    actual_num = get_available_slots(app, "anything", max_length)
    assert actual_num == max_length - queue_length
    assert mock.called


def test_setup_log_handler(capsys):
    log_level = logging.DEBUG

    setup_log_handler(loglevel=log_level)

    loggers = [
        ("celery", logging.INFO),
        ("amqp", log_level),
        ("urllib3", logging.WARNING),
        ("azure.core.pipeline.policies.http_logging_policy", logging.WARNING),
        ("swh", log_level),
        ("celery.task", log_level),
    ]

    for module, log_level in loggers:
        logger = logging.getLogger(module)
        assert logger.getEffectiveLevel() == log_level

    # exceptions are caught and displayed by the setup_log_handler function
    # as celery eats tracebacks in signal handler, check no traceback was
    # displayed then
    assert "Traceback" not in capsys.readouterr().err


def test_setup_log_handler_with_env_configuration(capsys, monkeypatch, datadir):
    logging_config_path = datadir / "logging-config.yaml"
    with monkeypatch.context() as m:
        m.setenv("SWH_LOG_CONFIG", logging_config_path)
        from os import environ

        assert environ["SWH_LOG_CONFIG"] == str(logging_config_path)
        setup_log_handler()

    # exceptions are caught and displayed by the setup_log_handler function
    # as celery eats tracebacks in signal handler, check no traceback was
    # displayed then
    assert "Traceback" not in capsys.readouterr().err

    # Reading the actual config and checking it's actually applied
    with open(logging_config_path, "r") as f:
        config_d = yaml.safe_load(f.read())

    for module, config in config_d["loggers"].items():
        logger = logging.getLogger(module)

        assert logger.getEffectiveLevel() == logging.getLevelName(config["level"])



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
from testing.postgresql import find_program

from .test_scheduler import TestScheduler


def _has_initdb():
    try:
        find_program("initdb", ["bin"])
    except RuntimeError:
        return False
    else:
        return True


@pytest.fixture
def swh_scheduler_class():
    return "temporary"


@pytest.fixture
def swh_scheduler_config():
    return {}


pytestmark = pytest.mark.skipif(
    not _has_initdb(), reason="initdb executable is missing"
)


class TestTemporaryScheduler(TestScheduler):
    pass



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import time

import click

from . import cli


@cli.group("simulator")
def simulator():
    """Scheduler simulator."""
    pass


@simulator.command("fill-test-data")
@click.option(
    "--num-origins",
    "-n",
    type=int,
    default=100000,
    help="Number of listed origins to add",
)
@click.pass_context
def fill_test_data_command(ctx, num_origins):
    """Fill the scheduler with test data for simulation purposes."""
    from swh.scheduler.simulator import fill_test_data

    click.echo(f"Filling {num_origins:,} listed origins data...")
    start = time.monotonic()
    fill_test_data(ctx.obj["scheduler"], num_origins=num_origins)
    runtime = time.monotonic() - start
    click.echo(f"Completed in {runtime:.2f} seconds")


@simulator.command("run")
@click.option(
    "--scheduler",
    "-s",
    type=click.Choice(["task_scheduler", "origin_scheduler"]),
    default="origin_scheduler",
    help="Scheduler to simulate",
)
@click.option(
    "--policy",
    "-p",
    default="oldest_scheduled_first",
    help="Scheduling policy to simulate (only for origin_scheduler)",
)
@click.option("--runtime", "-t", type=float, help="Simulated runtime")
@click.option(
    "--plots/--no-plots",
    "-P",
    "showplots",
    help="Show results as plots (with plotille)",
)
@click.option(
    "--csv", "-o", "csvfile", type=click.File("w"), help="Export results in a CSV file"
)
@click.pass_context
def run_command(ctx, scheduler, policy, runtime, showplots, csvfile):
    """Run the scheduler simulator.

    By default, the simulation runs forever. You can cap the simulated runtime
    with the --runtime option, and you can always press Ctrl+C to interrupt the
    running simulation.

    'task_scheduler' is the "classic" task-based scheduler; 'origin_scheduler'
    is the new origin-visit-aware simulator. The latter uses --policy to decide
    which origins to schedule first based on information from listers.
    """
    from swh.scheduler.simulator import run

    policy = policy if scheduler == "origin_scheduler" else None
    report = run(
        scheduler=ctx.obj["scheduler"],
        scheduler_type=scheduler,
        policy=policy,
        runtime=runtime,
    )

    print(report.format(with_plots=showplots))
    if csvfile is not None:
        report.metrics_csv(csvfile)
